import functools
import json
import re
import typing as t
from urllib import parse as urlparse

import responses  # type: ignore

from abnosql.plugins.table.memory import get_table_count
from abnosql import table


KEY_ATTRS: t.Dict[str, t.List[str]] = {}
CRYPTO_ATTRS: t.Dict[str, t.List[str]] = {}


def set_keyattrs(key_attrs: t.Dict[str, t.List[str]]):
    global KEY_ATTRS
    KEY_ATTRS = key_attrs


def set_crypto_attrs(attrs: t.Dict[str, t.List[str]]):
    global CRYPTO_ATTRS
    CRYPTO_ATTRS = attrs


def mock_cosmos(f):

    def _get_key(headers, key_attrs, doc_id):
        hk = key_attrs[0]
        rk = key_attrs[1] if len(key_attrs) > 1 else None
        _part_keys = headers.get('x-ms-documentdb-partitionkey')
        if isinstance(_part_keys, str) and len(_part_keys) > 0:
            _part_val = json.loads(_part_keys)[0]
        else:
            _part_val = doc_id
        key = {
            hk: _part_val
        }
        if rk is not None:
            key[rk] = doc_id
        return key

    def _callback(request):
        path = urlparse.urlsplit(request.url).path
        headers = dict(request.headers)

        def _response(code=404, body=None, _headers=None):
            return (
                code, _headers or {}, json.dumps({
                    "Errors": [
                        "Resource Not Found. "
                        "Learn more: https://aka.ms/cosmosdb-tsg-not-found"
                    ]
                }) if code == 404
                else json.dumps(body) if body is not None
                else body
            )

        parts = [_ for _ in path.split('/') if _ != '']
        # print(f'REQ: {request.method} {path} H: {headers} B: {request.body}')

        # required for CosmosClient
        if request.method == 'GET' and path == '/':
            return _response(
                200, {
                    "userConsistencyPolicy": {
                        "defaultConsistencyLevel": "Session"
                    }
                }
            )

        if len(parts) < 4 or parts[0] != 'dbs':
            return _response(404)

        table_name = parts[3]
        global KEY_ATTRS
        key_attrs = KEY_ATTRS.get(table_name)
        if key_attrs is None:
            return _response(404)

        # use memory table to mock cosmos
        global CRYPTO_ATTRS
        crypto_attrs = CRYPTO_ATTRS.get(table_name)
        config = {'key_attrs': key_attrs}
        if crypto_attrs is not None:
            config['crypto'] = {
                'attrs': crypto_attrs,
                'key_attrs': key_attrs
            }
        tb = table(
            table_name,
            config=config,
            database='memory'
        )

        # /dbs/{database}/colls/{table}/docs/{docid}
        if len(parts) == 6 and parts[-2] == 'docs':
            key = _get_key(headers, key_attrs, parts[-1])
            if request.method == 'GET':
                item = tb.get_item(**key)
                if item is not None:
                    return _response(200, item)
            elif request.method == 'DELETE':
                tb.delete_item(**key)
                return _response(204, None)

        # /dbs/{database}/colls/{table}/docs
        elif len(parts) == 5 and parts[-1] == 'docs':
            if request.method == 'POST':
                is_query = headers.get('x-ms-documentdb-isquery') == 'true'
                item = json.loads(request.body)
                if is_query is True:
                    items = tb.query_sql(
                        item['query'],
                        {
                            _['name']: _['value']
                            for _ in item.get('parameters', {})
                        }
                    )
                    headers = {
                        'x-ms-resource-usage': 'documentsCount=%s' % (
                            get_table_count(table_name)
                        )
                    }
                    return _response(
                        200, {'Documents': items}, headers
                    )
                else:
                    tb.put_item(item)
                return _response(201, None)

        # upsert_item() reads the collection
        # /dbs/{database}/colls/{table}
        elif len(parts) == 4 and parts[-2] == 'colls':
            if request.method == 'GET':
                return _response(200, {})

        return _response(404)

    @functools.wraps(f)
    def decorated(*args, **kwargs):
        for method in ['GET', 'POST', 'DELETE', 'PUT']:
            responses.add_callback(
                getattr(responses, method),
                re.compile(r'^https://.*.documents.azure.(com|cn).*'),
                _callback
            )
        return f(*args, **kwargs)
    return decorated




class NoSQLException(Exception):
    pass  # noqa


class NotFoundException(NoSQLException):
    pass  # noqa


class ConfigException(NoSQLException):
    pass  # noqa


class PluginException(NoSQLException):
    pass  # noqa


class ValidationException(NoSQLException):
    pass  # noqa



from abnosql.mocks.mock_azure_kms import mock_azure_kms
from abnosql.mocks.mock_cosmos import mock_cosmos
from abnosql.mocks.mock_dynamodbx import mock_dynamodbx


__all__ = [  # type: ignore
    mock_azure_kms,
    mock_dynamodbx,
    mock_cosmos
]



#!/usr/bin/env python
# mypy: ignore-errors
import json
import os

import click
from tabulate import tabulate  # type: ignore

from abnosql import table as _table


def dump(obj):
    if isinstance(obj, list):
        print(tabulate(obj, headers='keys'))
    else:
        print(json.dumps(obj, indent=2))


def get_key(partition_key, id_key=None):
    key = {}
    if len(partition_key.split('=')) != 2:
        raise ValueError('--partition-key format is name=value')
    vals = partition_key.split('=')
    key = {
        vals[0]: vals[1]
    }
    if id_key is not None:
        if len(id_key.split('=')) != 2:
            raise ValueError('--id-key format is name=value')
        vals = id_key.split('=')
        key[vals[0]] = vals[1]
    return key


def parse_dict_arg(arg):
    d = {}
    if arg is None:
        return None
    for val in arg.split(','):
        parts = [_.strip() for _ in val.split('=') if _.strip() != '']
        if len(parts) == 2:
            d[parts[0]] = parts[1]
    return d


def get_obj(argname, _obj, file_only=False):
    obj = None
    if os.path.isfile(_obj):
        obj = json.loads(open(_obj, 'r').read())
    elif file_only is True:
        raise ValueError(f'{argname} must be json file')
    elif _obj.startswith('{'):
        obj = json.loads(_obj)
    elif '=' in _obj:
        obj = parse_dict_arg(_obj)
    return obj


def get_config(obj):
    config_file = os.environ.get('ABNOSQL_CONFIG')
    if obj is None:
        if config_file is not None and os.path.isfile(config_file):
            obj = config_file
        else:
            return None
    return get_obj('--config', obj, True)


@click.group()
def cli():
    pass


@click.command()
@click.argument('table')
@click.argument('partition-key')
@click.option('--id-key', '-k')
@click.option('--database', '-d')
@click.option('--config', '-c')
def get_item(table, partition_key, id_key, database, config):
    tb = _table(table, get_config(config), database=database)
    dump(tb.get_item(**get_key(partition_key, id_key)))


@click.command()
@click.argument('table')
@click.argument('item')
@click.option('--database', '-d')
@click.option('--config', '-c')
def put_item(table, item, database, config):
    tb = _table(table, get_config(config), database=database)
    tb.put_item(get_obj('item', item))
    print('created')


@click.command()
@click.argument('table')
@click.argument('items')
@click.option('--database', '-d')
@click.option('--config', '-c')
def put_items(table, item, database, config):
    tb = _table(table, get_config(config), database=database)
    tb.put_item(get_obj('--items', item, True))
    print('created')


@click.command()
@click.argument('table')
@click.argument('partition-key')
@click.option('--id-key', '-i')
@click.option('--database', '-d')
@click.option('--config', '-c')
def delete_item(table, partition_key, id_key, database, config):
    tb = _table(table, get_config(config), database=database)
    tb.delete_item(**get_key(partition_key, id_key))
    print('deleted')


@click.command()
@click.argument('table')
@click.argument('partition-key')
@click.option('--id-key', '-k')
@click.option('--filters', '-f')
@click.option('--database', '-d')
@click.option('--config', '-c')
def query(table, partition_key, id_key, filters, database, config):
    tb = _table(table, get_config(config), database=database)
    dump(tb.query(
        key=get_key(partition_key, id_key),
        filters=parse_dict_arg(filters)
    )['items'])


@click.command()
@click.argument('table')
@click.argument('statement')
@click.option('--parameters', '-p')
@click.option('--database', '-d')
@click.option('--config', '-c')
def query_sql(table, statement, parameters, database, config):
    tb = _table(table, get_config(config), database=database)
    dump(tb.query_sql(
        statement=statement,
        parameters=parameters,
        limit=1
    )['items'])


cli.add_command(get_item)
cli.add_command(put_item)
cli.add_command(put_items)
cli.add_command(delete_item)
cli.add_command(query)
cli.add_command(query_sql)

if __name__ == '__main__':
    cli()



from base64 import b64decode
from base64 import b64encode
import functools
import os
import typing as t

import abnosql.exceptions as ex
from abnosql.kms import get_keys
from abnosql.kms import KmsBase
from abnosql.plugin import PM


try:
    import aws_encryption_sdk  # type: ignore
    from aws_encryption_sdk import CommitmentPolicy  # type: ignore
    from botocore.exceptions import ClientError  # type: ignore
    from botocore.exceptions import NoCredentialsError  # type: ignore
    from botocore.session import Session  # type: ignore
except ImportError:
    MISSING_DEPS = True


AWS_DEFAULT_REGION = os.environ.get('AWS_DEFAULT_REGION', 'us-east-1')


def kms_ex_handler(raise_not_found: t.Optional[bool] = True):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except ClientError as e:
                code = e.response['Error']['Code']
                if raise_not_found and code in ['ResourceNotFoundException']:
                    raise ex.NotFoundException(e) from None
                elif code == 'UnrecognizedClientException':
                    raise ex.ConfigException(e) from None
                raise ex.ValidationException(e) from None
            except NoCredentialsError as e:
                raise ex.ConfigException(e) from None
            except Exception as e:
                raise ex.PluginException(e)
        return wrapper
    return decorator


class Kms(KmsBase):

    def __init__(
        self, pm: PM, config: t.Optional[dict] = None
    ) -> None:
        self.pm = pm
        self.config = config or {}
        self.session = self.config.get(
            'session', Session()
        )
        self.key_ids = self.config.get('key_ids', get_keys())
        if not isinstance(self.key_ids, list) or len(self.key_ids) == 0:
            raise ex.ConfigException('kms key_ids required')
        self.client = aws_encryption_sdk.EncryptionSDKClient(
            commitment_policy=CommitmentPolicy.REQUIRE_ENCRYPT_REQUIRE_DECRYPT
        )
        self.mkp = aws_encryption_sdk.StrictAwsKmsMasterKeyProvider(
            key_ids=self.key_ids,
            botocore_session=self.session
        )

    @kms_ex_handler()
    def encrypt(
        self, plaintext: str, context: t.Dict, key: t.Optional[bytes] = None
    ) -> str:
        # not using aws dynamodb encryption sdk in case in future
        # we want to use another aws database (eg postgres)
        ciphertext, _ = self.client.encrypt(
            source=plaintext,
            key_provider=self.mkp,
            encryption_context=context
        )
        return b64encode(ciphertext).decode()

    @kms_ex_handler()
    def decrypt(self, serialized: str, context: t.Dict) -> str:
        plaintext, header = self.client.decrypt(
            source=b64decode(serialized),
            key_provider=self.mkp
        )
        for k, v in header.encryption_context.items():
            if k == 'aws-crypto-public-key':
                continue
            if v != context.get(k):
                raise ex.ValidationException(
                    'context mismatch'
                )
        return plaintext.decode()
